/************************************************************************/
/* comm.c - Communication Plugin Architecture Source			*/
/************************************************************************/
/*
 *  (c)2003,2008,2009,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <comm/comm.h>
#include <comm/comm_portable.h>


#define PATH_SIZE	256

#define	COMM_GROUP_FLAG_LOOP		0x00000001

#ifdef HAVE_EPOLL
#define	EPOLL_TABLE_SIZE	512
#endif


LIBCOMM_API int comm_plugin_load (CommModule **cmp, char *path, Attribute *attribs)
{
	int rc;
	int len;
	int size;
	char envstr[PATH_SIZE];
	char *fullpath = NULL;
	CommModule *cm = NULL;
	if (cmp == NULL || path == NULL)
		return EINVAL;
	*cmp = NULL;
	cm = (CommModule *)calloc (1, sizeof (CommModule));
	if (cm == NULL)
		return ENOMEM;
	if (strstr(path, DIR_DELIMITER) != NULL) {
		cm->path = strdup (path);
		if (cm->path == NULL) {
			free (cm);
			return ENOMEM;
		}
	} else {
#ifndef WIN32
		len = strlen (PKGLIBDIR) + strlen (path) + strlen (SHLIBEXT) + 8;
#else
		rc = GetEnvironmentVariable("PROGRAMFILES", envstr, PATH_SIZE);
		if (rc == 0)
			return ENOSYS;
		len = strlen(envstr) + strlen(OS_SUBDIR) + strlen(path) + strlen(SHLIBEXT) + 8;
#endif
		fullpath = calloc (1, len);
		if (fullpath == NULL) {
			free (cm);
			return ENOMEM;
		}
#ifndef WIN32
		size = snprintf (fullpath,
				len,
				"%s%s%s%s",
				PKGLIBDIR,
				DIR_DELIMITER,
				path,
				SHLIBEXT);
#else
		size = snprintf (fullpath,
				len,
				"%s%s%s%s%s%s",
				envstr,
				DIR_DELIMITER,
				OS_SUBDIR,
				DIR_DELIMITER,
				path,
				SHLIBEXT);
#endif
		if (size == -1 || size >= len) {
			free (fullpath);
			free (cm);
			return ENOMEM;
		}
		cm->path = fullpath;
	}

	cm->module = dlopen (cm->path, RTLD_NOW);
	if (cm->module == NULL) {
		rc = errno;
		free (cm->path);
		free (cm);
		return rc;
	}
	cm->load = dlsym (cm->module, "comm_init");
	if (cm->load == NULL) {
		rc = errno;
		dlclose (cm->module);
		free (cm->path);
		free (cm);
		return rc;
	}
	cm->unload = dlsym (cm->module, "comm_cleanup");
	if (cm->load == NULL) {
		rc = errno;
		dlclose (cm->module);
		free (cm->path);
		free (cm);
		return rc;
	}
	cm->moddata = (CommModuleData *)calloc (1, sizeof (CommModuleData));
	if (cm->moddata == NULL) {
		dlclose (cm->module);
		free (cm->path);
		free (cm);
		return ENOMEM;
	}

	cm->moddata->vector = (CommVectors *)calloc (1, sizeof (CommVectors));
	if (cm->moddata->vector == NULL) {
		dlclose (cm->module);
		free (cm->moddata);
		free (cm->path);
		free (cm);
		return ENOMEM;
	}

	cm->moddata->attribute = attribs;
	rc = (*cm->load) (cm->moddata);
	if (rc != 0) {
		free (cm->moddata->vector);
		free (cm->moddata);
		dlclose (cm->module);
		free (cm->path);
		free (cm);
		return rc;
	}
	*cmp = cm;
	return 0;
}


LIBCOMM_API int comm_plugin_unload (CommModule **cmp)
{
	int rc;
	CommModule *cm = NULL;
	if (cmp == NULL || *cmp == NULL)
		return EINVAL;
	cm = *cmp;
	if (cm->unload == NULL)
		return EINVAL;
	rc = (*cm->unload) (cm->moddata);
	if (rc != 0)
		return rc;
	free (cm->moddata->vector);
	free (cm->moddata);
	dlclose (cm->module);
	free (cm->path);
	free (cm);
	*cmp = NULL;
	return 0;
}


LIBCOMM_API int comm_bind (CommModule *cm, CommChannel **ccp,
		SList *interfacelist, int port)
{
	int rc;
	CommChannel *cc = NULL;
	if (cm == NULL || cm->moddata == NULL || ccp == NULL)
		return EINVAL;
	*ccp = NULL;
	if (cm->moddata->vector->bind == NULL)
		return EINVAL;
	cc = (CommChannel *)calloc (1, sizeof (CommChannel));
	if (cc == NULL)
		return ENOMEM;
	cc->cmodule = cm;
	rc = (cm->moddata->vector->bind) (cc, interfacelist, port);
	if (rc != 0) {
		free (cc);
		return rc;
	}
	cm->refcount++;
	*ccp = cc;
	return 0;
}


LIBCOMM_API int comm_accept (CommChannel *cc, CommChannel **ccp, int commflags)
{
	int rc;
	CommChannel *ncc = NULL;
	if (cc == NULL || ccp == NULL)
		return EINVAL;
	*ccp = NULL;
	if (cc->cmodule == NULL || cc->cmodule->moddata == NULL)
		return EINVAL;
	if (cc->cmodule->moddata->vector->accept == NULL)
		return EINVAL;
	ncc = (CommChannel *)calloc (1, sizeof (CommChannel));
	if (ncc == NULL)
		return ENOMEM;
	ncc->cmodule = cc->cmodule;
	ncc->flags = commflags;
	rc = (cc->cmodule->moddata->vector->accept) (cc, ncc, commflags);
	if (rc != 0) {
		if ( !(ncc->flags & COMMFLAG_INACCEPT) ) {
			free (ncc);
			return rc;
		}
	}
	cc->cmodule->refcount++;
	*ccp = ncc;
	return rc;
}


LIBCOMM_API int comm_accept_continue (CommChannel *cc)
{
	int rc;
	if (cc == NULL)
		return EINVAL;
	if ( !(cc->flags & COMMFLAG_INACCEPT) )
		return EINVAL;
	if (cc->cmodule == NULL || cc->cmodule->moddata == NULL)
		return EINVAL;
	if (cc->cmodule->moddata->vector->accept_continue == NULL)
		return EINVAL;
	rc = (cc->cmodule->moddata->vector->accept_continue) (cc);
	return rc;
}


LIBCOMM_API int comm_connect (CommModule *cm, CommChannel **ccp,
		char *host, int port)
{
	int rc;
	CommChannel *cc = NULL;
	if (cm == NULL || ccp == NULL || host == NULL)
		return EINVAL;
	*ccp = NULL;
	if (cm->moddata== NULL || cm->moddata->vector->connect == NULL)
		return EINVAL;
	cc = (CommChannel *)calloc (1, sizeof (CommChannel));
	if (cc == NULL)
		return ENOMEM;
	cc->cmodule = cm;
	rc = (cm->moddata->vector->connect) (cc, host, port);
	if (rc != 0) {
		free (cc);
		return rc;
	}
	cm->refcount++;
	*ccp = cc;
	return 0;
}


LIBCOMM_API int comm_read (CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->cmodule == NULL || cc->cmodule->moddata == NULL)
		return EINVAL;
	rc = (*cc->cmodule->moddata->vector->read) (cc, buffer, bytes);
	return rc;
}


LIBCOMM_API int comm_write (CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->cmodule == NULL || cc->cmodule->moddata == NULL)
		return EINVAL;
	rc = (*cc->cmodule->moddata->vector->write) (cc, buffer, bytes);
	return rc;
}


LIBCOMM_API int comm_close (CommChannel **ccp, int commflags)
{
	int rc;
	CommChannel *cc = NULL;
	if (ccp == NULL || *ccp == NULL)
		return EINVAL;
	cc = *ccp;
	if (cc->cmodule == NULL || cc->cmodule->moddata == NULL)
		return EINVAL;
	rc = (*cc->cmodule->moddata->vector->close) (cc, commflags);
	if (rc != 0)
		return rc;
	cc->cmodule->refcount--;
	free (cc);
	*ccp = NULL;
	return 0;
}


LIBCOMM_API int comm_get_flags (CommChannel *cc, int *commflags)
{
	if (cc == NULL || commflags == NULL)
		return EINVAL;
	*commflags = cc->flags;
	return 0;
}


LIBCOMM_API int comm_set_flags (CommChannel *cc, int commflags)
{
	if (cc == NULL)
		return EINVAL;
	cc->flags = commflags;
	return 0;
}


LIBCOMM_API int comm_get_fd (CommChannel *cc, int *fdp)
{
	int		rc;
	if (cc == NULL || fdp == NULL)
		return EINVAL;
	if (cc->cmodule == NULL || cc->cmodule->moddata == NULL)
		return EINVAL;
	rc = (*cc->cmodule->moddata->vector->get_fd) (cc, fdp);
	return rc;
}


LIBCOMM_API int comm_get_ioneeds (CommChannel *cc, int *ioneeds)
{
	if (cc == NULL || ioneeds == NULL)
		return EINVAL;
	*ioneeds = cc->ioneeds;
	return 0;
}


LIBCOMM_API int comm_set_ioneeds (CommChannel *cc, int ioneeds)
{
#ifdef HAVE_EPOLL
	int			rc;
	struct epoll_event	epevent;
	CommGroupItem		*cgi = NULL;
#endif
	if (cc == NULL)
		return EINVAL;
	cc->ioneeds = ioneeds;
#ifdef HAVE_EPOLL
	cgi = (CommGroupItem *)cc->cgdata;
	epevent.data.ptr = cgi;
	epevent.events = 0;
	if (ioneeds & COMM_IO_READ)
		epevent.events |= EPOLLIN;
	if (ioneeds & COMM_IO_WRITE)
		epevent.events |= EPOLLOUT;
	if (ioneeds & COMM_IO_EXCEPTION)
		epevent.events |= (EPOLLERR | EPOLLHUP);
	rc = epoll_ctl (cgi->cg->epfd, EPOLL_CTL_MOD, cgi->ccfd, &epevent);
	if (rc == -1) {
		rc = errno;
		free (cgi);
		return rc;
	}
#endif
	return 0;
}


