/************************************************************************/
/* comm_group.c - Communication Plugin Architecture Group Source	*/
/************************************************************************/
/*
 *  (c)2003,2009,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include <comm/comm.h>
#include <comm/comm_portable.h>


#define	COMM_GROUP_FLAG_LOOP		0x00000001

#ifdef HAVE_EPOLL
#define	EPOLL_TABLE_SIZE	512
#endif

typedef struct  _CommGroupData	CommGroupData;
#ifdef HAVE_EPOLL
struct	_CommGroupData {
	int		epfd;
	int		pipe[2];
};
#endif
#ifdef WIN32
struct	_CommGroupData {
	int		items;
	HANDLE		pipe;
	int		reason;
};
#endif


LIBCOMM_API int comm_group_new (CommGroup **cgp, CommIOCallback cb)
{
	int			rc;
#ifdef HAVE_EPOLL
	struct epoll_event	epevent;
#endif
	CommGroupData	*cgd = NULL;
	CommGroup	*cg = NULL;
	if (cgp == NULL)
		return EINVAL;
	*cgp = NULL;
	cg = calloc (1, sizeof (CommGroup));
	if (cg == NULL)
		return ENOMEM;
	cg->callback = cb;
	cgd = calloc (1, sizeof (CommGroupData));
	if (cgd == NULL) {
		free (cg);
		return ENOMEM;
	}
	cg->sdata = cgd;
#ifndef WIN32
	rc = pipe (cgd->pipe);
	if (rc == -1) {
		rc = errno;
		free (cgd);
		free (cg);
		return rc;
	}
#else
	cgd->pipe = CreateEvent (NULL, FALSE, FALSE, NULL);
	if (cgd->pipe == NULL) {
		rc = GetLastError ();
		free (cgd);
		free (cg);
		return rc;
	}
#endif
#ifdef HAVE_EPOLL
	cgd->epfd = epoll_create (EPOLL_TABLE_SIZE);
	if (cgd->epfd == -1) {
		rc = errno;
		close (cgd->pipe[0]);
		close (cgd->pipe[1]);
		free (cgd);
		free (cg);
		return rc;
	}
	epevent.data.fd = cgd->pipe[COMM_READ_FD];
	epevent.events = EPOLLIN;
	rc = epoll_ctl (cgd->epfd,
			EPOLL_CTL_ADD,
			cgd->pipe[COMM_READ_FD],
			&epevent);
	if (rc == -1) {
		rc = errno;
		close (cgd->epfd);
		close (cgd->pipe[0]);
		close (cgd->pipe[1]);
		free (cgd);
		free (cg);
		return rc;
	}
#endif
	*cgp = cg;
	return 0;
}


LIBCOMM_API int comm_group_free (CommGroup **cgp)
{
	int		rc;
	CommGroup	*cg = NULL;
	CommGroupItem	*cgi = NULL;
	CommGroupData	*cgd = NULL;
	SList		*litem = NULL;
	if (cgp == NULL || *cgp == NULL)
		return EINVAL;
	cg = *cgp;
	cgd = (CommGroupData *)cg->sdata;
	litem = cg->cclist;
	while (litem != NULL) {
		cgi = slist_get_data (litem);
		free (cgi);
		litem = slist_get_next (litem);
	}
	rc = slist_free (&cg->cclist);
#ifndef WIN32
	close (cgd->pipe[0]);
	close (cgd->pipe[1]);
#ifdef HAVE_EPOLL
	close (cgd->epfd);
#endif
#else
	CloseHandle (cgd->pipe);
#endif
	free (cgd);
	free (cg);
	*cgp = NULL;
	return 0;
}


LIBCOMM_API int comm_group_set_option (CommGroup *cg, int option, unsigned long value)
{
	if (cg == NULL)
		return EINVAL;
	switch (option) {
		case COMM_OPTION_TIMEOUT:
			if (value == 0) {
				cg->flags -= (cg->flags & COMM_OPTION_TIMEOUT);
				cg->timeout = 0;
			} else {
				cg->flags |= COMM_OPTION_TIMEOUT;
				cg->timeout = value;
			}
			break;
		default:
			return EINVAL;
	}
	return 0;
}


LIBCOMM_API int comm_group_get_option (CommGroup *cg, int option, unsigned long *value)
{
	if (cg == NULL)
		return EINVAL;
	switch (option) {
		case COMM_OPTION_TIMEOUT:
			*value = cg->timeout;
			break;
		default:
			return EINVAL;
	}
	return 0;
}


LIBCOMM_API int comm_group_add (CommGroup *cg, CommChannel *cc, void *ccarg, unsigned int iotype)
{
	int			rc;
	int			cfd;
#ifdef HAVE_EPOLL
	struct epoll_event	epevent;
#endif
	CommGroupItem		*cgi = NULL;
	CommGroupData		*cgd = NULL;
	if (cg == NULL || cc == NULL)
		return EINVAL;
	cgd = (CommGroupData *)cg->sdata;
	rc = comm_get_fd (cc, &cfd);

	/*
	 * Just set O_NONBLOCK (only valid socket flag)
	*/
	rc = fcntl (cfd, F_SETFL, O_NONBLOCK);
	if (rc != 0)
		return errno;

#ifdef WIN32
	if (cgd->items >= MAXIMUM_WAIT_OBJECTS)
		return ENOSPC;
	cc->sdata = CreateEvent (NULL, TRUE, FALSE, NULL);
	if (cc->sdata == NULL) {
		cc->sdata = NULL;
		return WSAGetLastError();
	}
#endif
	cgi = calloc (1, sizeof (CommGroupItem));
	if (cgi == NULL)
		return ENOMEM;
	cgi->cg = cg;
	cgi->cc = cc;
	cgi->ccarg = ccarg;
	cgi->ccfd = cfd;
	cgi->iotype = iotype;
	/* TODO: lock mutex? */
	slist_item_add (&cg->cclist, cgi);
#ifdef WIN32
	cgd->items++;
#endif
	/* TODO: unlock mutex? */
#ifdef HAVE_EPOLL
	epevent.data.ptr = cgi;
	epevent.events = 0;
	if (iotype & COMM_IO_READ)
		epevent.events |= EPOLLIN;
	if (iotype & COMM_IO_WRITE)
		epevent.events |= EPOLLOUT;
	if (iotype & COMM_IO_EXCEPTION)
		epevent.events |= (EPOLLERR | EPOLLHUP);
	rc = epoll_ctl (cgd->epfd, EPOLL_CTL_ADD, cfd, &epevent);
	if (rc == -1) {
		rc = errno;
		free (cgi);
		return rc;
	}
#endif
#ifndef WIN32
	cc->cgdata = cgi;
	rc = write (cgd->pipe[COMM_WRITE_FD], "R", 1);
#else
	rc = SetEvent (cgd->pipe);
#endif
	return 0;
}


#ifdef	DISABLED_FUNCTION
LIBCOMM_API int comm_group_remove (CommGroup *cg, CommChannel *cc)
{
	int		rc;
	int		cfd;
	CommGroupItem	*cgi = NULL;
	CommGroupData	*cgd = NULL;
	SList		*litem = NULL;

	if (cg == NULL || cc == NULL)
		return EINVAL;
	cgd = (CommGroupData *)cg->sdata;
	/* TODO: lock mutex? */
	litem = cg->cclist;
	while (litem != NULL) {
		cgi = slist_get_data (litem);
		if (cgi->cc == cc) {
			slist_item_remove (&cg->cclist, cgi);
#ifdef WIN32
			cgd->items--;
#endif
			/* TODO: unlock mutex? */
			if (! (cc->flags & COMMFLAG_SERVER)) {
				cfd = cgi->ccfd;
				/* If you do this, some plugins could 
				 * block forever in close()
				fdopts = fcntl (cfd, F_GETFL);
				fdopts -= fdopts & O_NONBLOCK;
				rc = fcntl (cfd, F_SETFL, fdopts);
				*/
			}
#ifdef WIN32
			if (cc->sdata != NULL) {
				CloseHandle(cc->sdata);
				cc->sdata = NULL;
			}
#endif
			free (cgi);
#ifdef HAVE_EPOLL
			rc = epoll_ctl (cgd->epfd, EPOLL_CTL_DEL, cfd, NULL);
			if (rc == -1)
				return errno;
#endif
#ifndef WIN32
			rc = write (cgd->pipe[COMM_WRITE_FD], "R", 1);
#else
			rc = SetEvent (cgd->pipe);
#endif
			return 0;
		}
		litem = slist_get_next (litem);
	}
	/* TODO: unlock mutex? */
	return ENOENT;
}
#endif


LIBCOMM_API int comm_group_remove (CommGroup *cg, CommChannel *cc)
{
	int		rc;
	int		cfd;
	CommGroupData	*cgd = NULL;
	CommGroupItem	*cgi = NULL;
#ifndef HAVE_EPOLL
	CommGroupItem	*cgis = NULL;
	SList		*litem = NULL;
#endif

	if (cg == NULL || cc == NULL)
		return EINVAL;
	cgd = (CommGroupData *)cg->sdata;

	/* TODO: lock mutex? */

	/*
	 * Get the CommGroupItem for the given CommChannel
	*/
#ifdef HAVE_EPOLL
	if (cc->cgdata == NULL)
		return EINVAL;
	cgi = (CommGroupItem *)cc->cgdata;
#else
	cgi = NULL;
	litem = cg->cclist;
	while (litem != NULL) {
		cgis = slist_get_data (litem);
		if (cgis->cc == cc) {
			cgi = cgis;
			break;
		}
		litem = slist_get_next (litem);
	}
	if (cgi == NULL)
		return ENOENT;
#endif

	slist_item_remove (&cg->cclist, cgi);
#ifdef WIN32
	cgd->items--;
#endif
	/* TODO: unlock mutex? */

#ifdef HAVE_EPOLL
	cfd = cgi->ccfd;
	rc = epoll_ctl (cgd->epfd, EPOLL_CTL_DEL, cfd, NULL);
	if (rc == -1)
		return errno;
#endif

#ifndef WIN32
	rc = write (cgd->pipe[COMM_WRITE_FD], "R", 1);
#else
	rc = SetEvent (cgd->pipe);
#endif

	free (cgi);
	return 0;
}


#ifdef HAVE_EPOLL
LIBCOMM_API int comm_group_loop (CommGroup *cg)
{
	int			rc;
	char			c;
	int			i;
	int			count;
	int			timeout;
	struct epoll_event	epevent[EPOLL_TABLE_SIZE];
	CommGroupItem		*cgi = NULL;
	CommGroupData		*cgd = NULL;
	if (cg == NULL)
		return EINVAL;
	cgd = (CommGroupData *)cg->sdata;
	cg->flags |= COMM_GROUP_FLAG_LOOP;
	while (1) {
		if ( !(cg->flags & COMM_GROUP_FLAG_LOOP) )
			break;
		if (cg->timeout == 0)
			timeout = -1;
		else
			timeout = cg->timeout / 1000;
		rc = epoll_wait (cgd->epfd,
				epevent,
				EPOLL_TABLE_SIZE,
				timeout);
		if (rc == -1) {
			return errno;
		} else if (rc == 0) {
			/* TODO: handle any timeout events/callbacks */
			if (cg->timeoutcallback != NULL) {
				rc = (*cg->timeoutcallback)(cg, cg->timeoutarg);
				if (rc != 0)
					return rc;
			}
		} else {
			count = rc;
			for (i=0 ; i<count ; i++) {
				if (epevent[i].data.fd == cgd->pipe[COMM_READ_FD]) {
					rc = read (cgd->pipe[COMM_READ_FD], &c, 1);
					if (c == 'R')
						continue;
					else if (c == 'E')
						return 0;
					continue;
				}
				cgi = epevent[i].data.ptr;
				if (epevent[i].events & EPOLLIN)
					rc = (*cg->callback)(cgi->cc,
							COMM_IO_READ,
							cgi->ccarg);
				else if (epevent[i].events & EPOLLOUT)
					rc = (*cg->callback)(cgi->cc,
							COMM_IO_WRITE,
							cgi->ccarg);
				else if (epevent[i].events & EPOLLERR)
					rc = (*cg->callback)(cgi->cc,
							COMM_IO_EXCEPTION,
							cgi->ccarg);
				else if (epevent[i].events & EPOLLHUP)
					rc = (*cg->callback)(cgi->cc,
							COMM_IO_EXCEPTION,
							cgi->ccarg);
			}
		}
	}
	return 0;
}
#endif


#ifdef HAVE_SELECT
LIBCOMM_API int comm_group_loop (CommGroup *cg)
{
	int			rc;
	char			c;
	int			maxfd;
	fd_set			rfds;
	fd_set			wfds;
	fd_set			efds;
	struct timeval		tv;
	struct timeval		*tvp = NULL;
	SList			*litem = NULL;
	CommGroupItem		*cgi = NULL;
	CommGroupDat		*cgd = NULL;
	if (cg == NULL)
		return EINVAL;
	cgd = (CommGroupData *)cg->sdata;
	cg->flags |= COMM_GROUP_FLAG_LOOP;
	while (1) {
		if ( !(cg->flags & COMM_GROUP_FLAG_LOOP) )
			break;
		maxfd = 0;
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_ZERO(&efds);
		FD_SET(cgd->pipe[COMM_READ_FD], &rfds);
		litem = cg->cclist;
		while (litem != NULL) {
			cgi = slist_get_data (litem);
			if (cgi->iotype & COMM_IO_READ) {
				FD_SET(cgi->ccfd, &rfds);
				if (cgi->ccfd > maxfd)
					maxfd = cgi->ccfd;
			}
			if ((cgi->iotype & COMM_IO_WRITE) ||
			    (cgi->cc->ioneeds & COMM_IO_WRITE)) {
				FD_SET(cgi->ccfd, &wfds);
				if (cgi->ccfd > maxfd)
					maxfd = cgi->ccfd;
			}
			if (cgi->iotype & COMM_IO_EXCEPTION) {
				FD_SET(cgi->ccfd, &efds);
				if (cgi->ccfd > maxfd)
					maxfd = cgi->ccfd;
			}
			litem = slist_get_next (litem);
		}
		maxfd++;
		if (cg->timeout == 0) {
			tvp = NULL;
		} else {
			tv.tv_sec = cg->timeout / 1000000;
			tv.tv_usec = cg->timeout - (tv.tv_sec * 1000000);
			tvp = &tv;
		}
		rc = select (maxfd, &rfds, &wfds, &efds, tvp);
		if (rc == -1) {
			rc = errno;
			FD_ZERO(&rfds);
			FD_ZERO(&wfds);
			FD_ZERO(&efds);
			return rc;
		} else if (rc == 0) {
			/* TODO: handle any timeout events/callbacks */
			if (cg->timeoutcallback != NULL) {
				rc = (*cg->timeoutcallback)(cg, cg->timeoutarg);
				if (rc != 0)
					return rc;
			}
		} else {
			if (rc == ENODATA) {
				return rc;
			}
			c = 0;
			if (FD_ISSET(cgd->pipe[COMM_READ_FD], &rfds)) {
				rc = read (cgd->pipe[COMM_READ_FD], &c, 1);
				if (c == 'R')
					continue;
				else if (c == 'E')
					break;
		/*		continue; */
			}
			litem = cg->cclist;
			while (litem != NULL) {
				cgi = slist_get_data (litem);
				if (cgi != NULL) {
					if (FD_ISSET(cgi->ccfd, &rfds))
						rc = (*cg->callback)(cgi->cc,
								COMM_IO_READ,
								cgi->ccarg);
						if (rc == ENODATA){
							break;
						}
					if (FD_ISSET(cgi->ccfd, &wfds))
						rc = (*cg->callback)(cgi->cc,
								COMM_IO_WRITE,
								cgi->ccarg);
						if (rc == ENODATA){
							break;
						}
					if (FD_ISSET(cgi->ccfd, &efds))
						rc = (*cg->callback)(cgi->cc,
								COMM_IO_EXCEPTION,
								cgi->ccarg);
						if (rc == ENODATA){
							break;
						}
				}
				litem = slist_get_next (litem);
			}
		}
	}
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	FD_ZERO(&efds);
	return 0;
}
#endif


#ifdef WIN32
LIBCOMM_API int comm_group_loop (CommGroup *cg)
{
	int			maxfd;
	uint32_t		rc;
	uint32_t		events;
	uint32_t		msecs;
	HANDLE			hlist[MAXIMUM_WAIT_OBJECTS];
	CommGroupItem		*cgilist[MAXIMUM_WAIT_OBJECTS];
	WSANETWORKEVENTS	nevent;
	SList			*litem = NULL;
	CommGroupItem		*cgi = NULL;
	CommGroupData		*cgd = NULL;
	if (cg == NULL)
		return EINVAL;
	cgd = (CommGroupData *)cg->sdata;
	cg->flags |= COMM_GROUP_FLAG_LOOP;
	while (1) {
		if ( !(cg->flags & COMM_GROUP_FLAG_LOOP) )
			break;
		maxfd = 0;

		/* Add the loop event */
		hlist[maxfd] = cgd->pipe;
		maxfd++;

		/* add each socket */
		litem = cg->cclist;
		while (litem != NULL && maxfd < MAXIMUM_WAIT_OBJECTS) {
			events = 0;
			cgi = slist_get_data (litem);
			if (cgi->iotype & COMM_IO_READ) {
				events |= FD_READ;
			}
			if ((cgi->iotype & COMM_IO_WRITE) ||
			    (cgi->cc->ioneeds & COMM_IO_WRITE)) {
				events |= FD_WRITE;
			}
			if (cgi->iotype & COMM_IO_EXCEPTION) {
				events |= FD_CLOSE;
			}
			if (cgi->cc->sdata != NULL) {
				rc = WSAEventSelect ((SOCKET)cgi->ccfd,
						(WSAEVENT)cgi->cc->sdata,
						events);
			}
			hlist[maxfd] = cgi->cc->sdata;
			cgilist[maxfd] = cgi;
			maxfd++;
			litem = slist_get_next (litem);
		}
		if (cg->timeout == 0) {
			msecs = INFINITE;
		} else {
			msecs = cg->timeout / 1000;
		}
		rc = WaitForMultipleObjects (maxfd,
					hlist,
					FALSE,
					msecs);
		if (rc == WAIT_FAILED) {
			rc = GetLastError();
			return rc;
		} else if (rc == WAIT_TIMEOUT) {
			/* TODO: handle any timeout events/callbacks */
			if (cg->timeoutcallback != NULL) {
				rc = (*cg->timeoutcallback)(cg, cg->timeoutarg);
				if (rc != 0)
					return rc;
			}
		} else {
			rc -= WAIT_OBJECT_0;
			if (rc == WAIT_OBJECT_0) {
				if (cgd->reason == 'R')
					continue;
				else if (cgd->reason == 'E')
					break;
		/*		continue; */
			} else {
				cgi = cgilist[rc];
				WSAEnumNetworkEvents (cgi->ccfd,
							cgi->cc->sdata,
							&nevent);
				if (nevent.lNetworkEvents & FD_READ) {
					rc = (*cg->callback)(cgi->cc,
							COMM_IO_READ,
							cgi->ccarg);
					if (rc == ENODATA){
						break;
					}
				}
				if (nevent.lNetworkEvents & FD_WRITE) {
					rc = (*cg->callback)(cgi->cc,
							COMM_IO_WRITE,
							cgi->ccarg);
					if (rc == ENODATA){
						break;
					}
				}
				if (nevent.lNetworkEvents & FD_CLOSE) {
					rc = (*cg->callback)(cgi->cc,
							COMM_IO_EXCEPTION,
							cgi->ccarg);
					if (rc == ENODATA){
						break;
					}
				}
			}
		}
	}
	return 0;
}
#endif



LIBCOMM_API int comm_group_exit_loop (CommGroup *cg)
{
	int		rc;
	CommGroupData	*cgd = NULL;
	if (cg == NULL)
		return EINVAL;
	cgd = (CommGroupData *)cg->sdata;
	cg->flags -= (cg->flags & COMM_GROUP_FLAG_LOOP);
#ifndef WIN32
	rc = write (cgd->pipe[COMM_WRITE_FD], "E", 1);
#else
	rc = SetEvent (cgd->pipe);
#endif
	return 0;
}


LIBCOMM_API int comm_group_restart_loop (CommGroup *cg)
{
	if (cg == NULL)
		return EINVAL;
	if (write (cg->pipefd[COMM_WRITE_FD], "R", 1) == -1)
		return errno;
	return 0;
}


LIBCOMM_API int comm_group_set_timeout_callback (CommGroup *cg,
                CommTimeoutCallback callback,
                void *arg)
{
	if (cg == NULL)
		return EINVAL;
	if (callback == NULL)
		return EINVAL;
	cg->timeoutcallback = callback;
	cg->timeoutarg = arg;
	return 0;
}


