/************************************************************************/
/* comm_gnutls.c - Communication Plugin Source (GNUTLS)			*/
/************************************************************************/
/*
 *  (c)2008,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <cutil/cutil.h>
#include <comm/comm.h>
#include <comm/comm_portable.h>

#include <gnutls/gnutls.h>

#ifdef GNUTLS_OLD_1_0
#define gnutls_session_t gnutls_session
#define gnutls_certificate_credentials_t gnutls_certificate_credentials
#define gnutls_dh_params_t gnutls_dh_params
#define gnutls_transport_ptr_t gnutls_transport_ptr
#endif

#include <gcrypt.h>
GCRY_THREAD_OPTION_PTHREAD_IMPL;


#define DH_BITS		1024



typedef struct _GnutlsChannelData	GnutlsChannelData;
struct _GnutlsChannelData {
	gnutls_session_t	sess;
	int			fd;
};


typedef struct _GnutlsModuleData	GnutlsModuleData;
struct _GnutlsModuleData {
	CommModuleData				*cmd;
	gnutls_certificate_credentials_t	xcred;
#ifdef GNUTLS_VERSION_2
	gnutls_priority_t			priocache;
#endif
	gnutls_dh_params_t			dhparams;
};


/* Function Prototypes */
int gnutls_comm_bind(CommChannel *cc, SList *interfacelist, int port);
int gnutls_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags);
int gnutls_comm_accept_continue(CommChannel *cc);
int gnutls_comm_connect(CommChannel *cc, char *host, int port);
int gnutls_comm_read(CommChannel *cc, char *buffer, int *bytes);
int gnutls_comm_write(CommChannel *cc, char *buffer, int *bytes);
int gnutls_comm_close(CommChannel *cc, int commflags);
int gnutls_comm_get_fd(CommChannel *cc, int *fdp);
int gnutls_map_return_code (int rc);

/* Global variables */
GnutlsModuleData	GMD;


/* Debug stuff */
/*
#define DEBUG		1
#ifdef DEBUG
#define ASSERTRC(rc) \
{ \
	if (rc != 0) { \
		fprintf (stderr, \
			"%s(%d): rc=%d\n", \
			__PRETTY_FUNCTION__, \
			__LINE__, \
			rc); \
	} \
}
#else
#define ASSERTRC(rc)
#endif
*/


int comm_init (CommModuleData *cmd)
{
	int		rc;
	char		*cafile = NULL;
	char		*certfile = NULL;
	char		*keyfile = NULL;
	char		*crlfile = NULL;
	if (cmd == NULL)
		return COMM_RC_INVALID_ARGUMENT;
        cafile = attribute_get_value (cmd->attribute, "SSLTrustedCAFile");
        certfile = attribute_get_value (cmd->attribute, "SSLCertificateFile");
        keyfile = attribute_get_value (cmd->attribute, "SSLPrivateKeyFile");
        crlfile = attribute_get_value (cmd->attribute, "SSLRevocationListFile");
        if (certfile == NULL || keyfile == NULL)
                return COMM_RC_INVALID_ARGUMENT;

	gcry_control (GCRYCTL_SET_THREAD_CBS,
			&gcry_threads_pthread);
	rc = gnutls_global_init ();
	if (rc != GNUTLS_E_SUCCESS)
		return gnutls_map_return_code(rc);
	gnutls_certificate_allocate_credentials (&GMD.xcred);
	if (cafile != NULL) {
		gnutls_certificate_set_x509_trust_file (GMD.xcred,
						cafile,
						GNUTLS_X509_FMT_PEM);
	}
	if (crlfile != NULL) {
		gnutls_certificate_set_x509_crl_file (GMD.xcred,
						crlfile,
						GNUTLS_X509_FMT_PEM);
	}
	if (certfile != NULL && keyfile != NULL) {
		gnutls_certificate_set_x509_key_file (GMD.xcred,
						certfile,
						keyfile,
						GNUTLS_X509_FMT_PEM);
	}

	cmd->name = "gnutls";
	cmd->vector->bind = gnutls_comm_bind;
	cmd->vector->accept = gnutls_comm_accept;
	cmd->vector->accept_continue = gnutls_comm_accept_continue;
	cmd->vector->connect = gnutls_comm_connect;
	cmd->vector->read = gnutls_comm_read;
	cmd->vector->write = gnutls_comm_write;
	cmd->vector->close = gnutls_comm_close;
	cmd->vector->get_fd = gnutls_comm_get_fd;
	GMD.cmd = cmd;
	return 0;
}


int comm_cleanup (CommModuleData *cmd)
{
	if (cmd == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	gnutls_certificate_free_credentials (GMD.xcred);
#ifdef GNUTLS_VERSION_2
	gnutls_priority_deinit (GMD.priocache);
#endif
	gnutls_global_deinit ();
	return 0;
}


int gnutls_comm_bind(CommChannel *cc, SList *interfacelist, int port)
{
	int rc;
	int sock;
	int socketopt;
	struct sockaddr_in sa;
	GnutlsChannelData	*gcd = NULL;
	if (cc == NULL || interfacelist == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	gcd = calloc (1, sizeof (GnutlsChannelData));
	if (gcd == NULL)
		return COMM_RC_MEMORY_ALLOCATION_FAILURE;
	sock = socket (PF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		rc = errno;
		free (gcd);
		return rc;
	}
	memset (&sa, 0, sizeof (struct sockaddr_in));
        sa.sin_family = AF_INET;
        sa.sin_port = htons (port);
        sa.sin_addr.s_addr = htonl (INADDR_ANY);
	socketopt = 1;
	setsockopt (sock,
		SOL_SOCKET,
		SO_REUSEADDR,
		&socketopt,
		sizeof (socketopt));
	rc = bind (sock,
		(struct sockaddr *)&sa,
		sizeof (struct sockaddr_in));
	if (rc == -1) {
		rc = errno;
		close (sock);
		free (gcd);
		return rc;
	}
	rc = listen (sock, 5);
	if (rc == -1) {
		rc = errno;
		close (sock);
		free (gcd);
		return rc;
	}
	gcd->fd = sock;
	cc->data = gcd;

	/* setup as a server */
	cc->flags |= COMMFLAG_SERVER;
	rc = gnutls_dh_params_init (&GMD.dhparams);
	if (rc != GNUTLS_E_SUCCESS) {
		close (sock);
		free (gcd);
		return gnutls_map_return_code(rc);
	}
	rc = gnutls_dh_params_generate2 (GMD.dhparams, DH_BITS);
	if (rc != GNUTLS_E_SUCCESS) {
		gnutls_dh_params_deinit (GMD.dhparams);
		close (sock);
		free (gcd);
		return gnutls_map_return_code(rc);
	}
#ifdef GNUTLS_VERSION_2
	gnutls_priority_init (&GMD.priocache, 
				"NORMAL",
				NULL);
#endif
	gnutls_certificate_set_dh_params (GMD.xcred,
					GMD.dhparams);
	return 0;
}



int gnutls_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags)
{
	int rc;
	int csock;
	int csalen;
	struct sockaddr_in csa;
	GnutlsChannelData	*gcd = NULL;
	GnutlsChannelData	*gcdc = NULL;
	if (cc == NULL || ncc == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	if (!(cc->flags & COMMFLAG_SERVER))
		return COMM_RC_INVALID_ARGUMENT;
	gcd = cc->data;
	if (gcd == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	csock = accept (gcd->fd,
			(struct sockaddr *)&csa,
			&csalen);
	if (csock == -1)
		return errno;

	/*
	 * Set the new socket non-blocking now
	 * Just set O_NONBLOCK (only valid socket flag)
	*/
	if (! (cc->flags & COMMFLAG_IO_BLOCK)) {
		rc = fcntl (csock, F_SETFL, O_NONBLOCK);
		if (rc != 0) {
			rc = errno;
			close (csock);
			return rc;
		}
	}

	gcdc = calloc (1, sizeof (GnutlsChannelData));
	if (gcdc == NULL) {
		close (csock);
		return COMM_RC_MEMORY_ALLOCATION_FAILURE;
	}
	gcdc->fd = csock;
	rc = gnutls_init (&gcdc->sess, GNUTLS_SERVER);
	if (rc != 0) {
		close (gcdc->fd);
		free (gcdc);
		return gnutls_map_return_code(rc);
	}
#ifdef GNUTLS_VERSION_2
	rc = gnutls_priority_set (gcdc->sess, GMD.priocache);
#else
	rc = gnutls_set_default_priority (gcdc->sess);
#endif
	ASSERTRC(rc);
	if (rc < 0) {
		gnutls_deinit (gcdc->sess);
		close (gcdc->fd);
		free (gcdc);
		return gnutls_map_return_code(rc);
	}
	gnutls_credentials_set (gcdc->sess,
				GNUTLS_CRD_CERTIFICATE,
				GMD.xcred);
/*
	gnutls_certificate_server_set_request (gcdc->sess, 
				GNUTLS_CERT_REQUEST);
*/
	gnutls_transport_set_ptr (gcdc->sess,
				(gnutls_transport_ptr_t) gcdc->fd);
	rc = gnutls_handshake (gcdc->sess);
	ASSERTRC(rc);
	if (rc != 0) {
		if (rc == GNUTLS_E_AGAIN || rc == GNUTLS_E_INTERRUPTED) {
			ncc->data = gcdc;
			ncc->flags |= COMMFLAG_INACCEPT;
			return gnutls_map_return_code(rc);
		} else {
			close (gcdc->fd);
			gnutls_deinit (gcdc->sess);
			free (gcdc);
			return gnutls_map_return_code(rc);
		}
	}
	ncc->data = gcdc;
	ncc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int gnutls_comm_accept_continue(CommChannel *cc)
{
	int			rc;
	GnutlsChannelData	*gcd = NULL;
	if (cc == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	gcd = cc->data;
	if (gcd == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	rc = gnutls_handshake (gcd->sess);
	ASSERTRC(rc);
	if (rc != 0) {
		if (rc == GNUTLS_E_AGAIN || rc == GNUTLS_E_INTERRUPTED) {
			return rc;
		}
		close (gcd->fd);
		gnutls_deinit (gcd->sess);
		free (gcd);
		return gnutls_map_return_code(rc);
	}
	cc->flags -= (cc->flags & COMMFLAG_INACCEPT);
	cc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int gnutls_comm_connect(CommChannel *cc, char *host, int port)
{
	int			rc;
	struct sockaddr_in	sa;
	const char		*err = NULL;
        struct hostent		*he = NULL;
	GnutlsChannelData	*gcd = NULL;
	if (cc == NULL || host == NULL || port < 1)
		return COMM_RC_INVALID_ARGUMENT;
        he = gethostbyname (host);
        if (he == NULL)
                return COMM_RC_NO_ENTRY;
	gcd = calloc (1, sizeof (GnutlsChannelData));
	if (gcd == NULL)
		return COMM_RC_MEMORY_ALLOCATION_FAILURE;
	memset (&sa, 0, sizeof (struct sockaddr_in));
        sa.sin_family = AF_INET;
        sa.sin_port = htons (port);
        sa.sin_addr.s_addr = *(long *)he->h_addr_list[0];
	gcd->fd = socket (PF_INET, SOCK_STREAM, 0);
	if (gcd->fd == -1) {
		rc = errno;
		free (gcd);
		return rc;
	}
	rc = connect (gcd->fd,
			(struct sockaddr *)&sa,
			sizeof (struct sockaddr_in));
	if (rc == -1) {
		rc = errno;
		close (gcd->fd);
		free (gcd);
		return rc;
	}

	rc = gnutls_init (&gcd->sess, GNUTLS_CLIENT);
	ASSERTRC(rc);
	if (rc != 0) {
		close (gcd->fd);
		free (gcd);
		return gnutls_map_return_code(rc);
	}
#ifdef GNUTLS_VERSION_2
	/* TODO: Use the priority cache for clients
	rc = gnutls_priority_set (gcdc->sess, GMD.priocache);
	*/
	rc = gnutls_priority_set_direct (gcd->sess,
				"PERFORMANCE",
				&err);
#else
	rc = gnutls_set_default_priority (gcd->sess);
#endif
	ASSERTRC(rc);
	if (rc != 0) {
		close (gcd->fd);
		free (gcd);
		return gnutls_map_return_code(rc);
	}

	gnutls_credentials_set (gcd->sess,
				GNUTLS_CRD_CERTIFICATE,
				GMD.xcred);

	gnutls_transport_set_ptr (gcd->sess,
				(gnutls_transport_ptr_t) gcd->fd);
	rc = gnutls_handshake (gcd->sess);
	ASSERTRC(rc);
	if (rc != 0) {
		close (gcd->fd);
		gnutls_deinit (gcd->sess);
		free (gcd);
		return gnutls_map_return_code(rc);
	}
	cc->data = gcd;
	cc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int gnutls_comm_read(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int len;
	int bytesleft;
	char *p = NULL;
	GnutlsChannelData	*gcd = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	if (cc->flags & COMMFLAG_SERVER)
		return COMM_RC_INVALID_ARGUMENT;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return COMM_RC_INVALID_ARGUMENT;
	gcd = cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = gnutls_record_recv (gcd->sess, p, bytesleft);
			if (rc == 0) {
				*bytes = len - bytesleft;
				return rc;
			} else if (rc < 0) {
				*bytes = len - bytesleft;
				return gnutls_map_return_code(rc);
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = gnutls_record_recv (gcd->sess, buffer, len);
		if (rc == 0)
			return COMM_RC_NO_DATA;
		else if (rc == -1)
			return errno;
		else if (rc < 0)
			return gnutls_map_return_code(rc);
		*bytes = rc;
	}
	return 0;
}


int gnutls_comm_write(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int len;
	int bytesleft;
	char *p = NULL;
	GnutlsChannelData	*gcd = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	if (cc->flags & COMMFLAG_SERVER)
		return COMM_RC_INVALID_ARGUMENT;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return COMM_RC_INVALID_ARGUMENT;
	gcd = cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = gnutls_record_send (gcd->sess, p, bytesleft);
			if (rc == 0) {
				*bytes = len - bytesleft;
				return rc;
			} else if (rc == -1) {
				*bytes = len - bytesleft;
				return errno;
			} else if (rc < 0) {
				*bytes = len - bytesleft;
				return gnutls_map_return_code(rc);
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = gnutls_record_send (gcd->sess, buffer, len);
		if (rc == -1)
			return errno;
		else if (rc < 0)
			return gnutls_map_return_code(rc);
		*bytes = rc;
	}
	return 0;
}


int gnutls_comm_close(CommChannel *cc, int commflags)
{
	int rc;
	GnutlsChannelData	*gcd = NULL;
	if (cc == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	gcd = cc->data;
	if (gcd == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	if (gcd->sess != NULL)
		gnutls_bye (gcd->sess, GNUTLS_SHUT_WR);
	if (gcd->fd != -1)
		rc = close (gcd->fd);
	gnutls_deinit (gcd->sess);
	free (gcd);
	return 0;
}


int gnutls_comm_get_fd(CommChannel *cc, int *fdp)
{
	GnutlsChannelData	*gcd = NULL;
	if (cc == NULL || fdp == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	gcd = cc->data;
	if (gcd == NULL)
		return COMM_RC_INVALID_ARGUMENT;
	*fdp = gcd->fd;
	return 0;
}



int gnutls_map_return_code (int rc)
{
	int	nrc;
	/* gnutls errors are negative numbers, fast-path anything else */
	if (rc >= 0)
		return rc;
	switch (rc) {
		case GNUTLS_E_AGAIN:
			nrc = COMM_RC_CALL_AGAIN;
			break;
		case GNUTLS_E_UNSUPPORTED_VERSION_PACKET:
			nrc = COMM_RC_UNSUPPORTED_PACKET_VERSION;
			break;
		case GNUTLS_E_UNEXPECTED_PACKET_LENGTH:
			nrc = COMM_RC_UNEXPECTED_PACKET_LENGTH;
			break;
		case GNUTLS_E_INVALID_SESSION:
			nrc = COMM_RC_INVALID_SESSION;
			break;
		case GNUTLS_E_FATAL_ALERT_RECEIVED:
			nrc = COMM_RC_FATAL_ALERT_RECEIVED;
			break;
		case GNUTLS_E_UNEXPECTED_PACKET:
			nrc = COMM_RC_UNEXPECTED_PACKET;
			break;
		case GNUTLS_E_WARNING_ALERT_RECEIVED:
			nrc = COMM_RC_WARNING_ALERT_RECEIVED;
			break;
		case GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET:
			nrc = COMM_RC_UNEXPECTED_HANDSHAKE_PACKET;
			break;
		case GNUTLS_E_DECRYPTION_FAILED:
			nrc = COMM_RC_DECRYPTION_FAILURE;
			break;
		case GNUTLS_E_DECOMPRESSION_FAILED:
			nrc = COMM_RC_DECOMPRESSION_FAILURE;
			break;
		case GNUTLS_E_COMPRESSION_FAILED:
			nrc = COMM_RC_COMPRESSION_FAILURE;
			break;
		case GNUTLS_E_INSUFFICIENT_CREDENTIALS:
			nrc = COMM_RC_INSUFFICIENT_CREDENTIALS;
			break;
		default:
			nrc = rc;
	}
	return nrc;
}

