/************************************************************************/
/* comm_openssl.c - Communication Plugin Source (OpenSSL)		*/
/************************************************************************/
/*
 *  (c)2003,2008,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>

#include <cutil/cutil.h>
#include <comm/comm.h>
#include <comm/comm_portable.h>


#define SSL_CIPHER_LIST		"ALL:!ADH:!EXP:!aNULL:!eNULL:!LOW:!SSLv2:!RC4:!SEED:+HIGH:+MEDIUM:@STRENGTH"
#define SSL_PROTOCOL_OPTIONS	(SSL_OP_ALL | SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3)


typedef struct _SslSysData	SslSysData;
struct _SslSysData {
	SSL_METHOD	*method;
	SSL_CTX		*context;
};

typedef struct _SslChannelData	SslChannelData;
struct _SslChannelData {
	SSL		*ssl;
	int		fd;
};


/* Function Prototypes */
int openssl_comm_bind(CommChannel *cc, SList *interfacelist, int port);
int openssl_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags);
int openssl_comm_accept_continue(CommChannel *cc);
int openssl_comm_connect(CommChannel *cc, char *host, int port);
int openssl_comm_read(CommChannel *cc, char *buffer, int *bytes);
int openssl_comm_write(CommChannel *cc, char *buffer, int *bytes);
int openssl_comm_close(CommChannel *cc, int commflags);
int openssl_comm_get_fd (CommChannel *cc, int *fdp);

/* Global variables */
SslSysData	ssl_sysdata;

int comm_init (CommModuleData *cmd)
{
	int rc;
	char *cafile = NULL;
	char *certfile = NULL;
	char *keyfile = NULL;
	if (cmd == NULL)
		return EINVAL;

	memset (&ssl_sysdata, 0, sizeof (SslSysData));
	cafile = attribute_get_value (cmd->attribute, "SSLTrustedCAFile");
	certfile = attribute_get_value (cmd->attribute, "SSLCertificateFile");
	keyfile = attribute_get_value (cmd->attribute, "SSLPrivateKeyFile");
	if (certfile == NULL || keyfile == NULL)
		return EINVAL;
	rc = SSL_library_init();
	SSL_load_error_strings();

	ssl_sysdata.method = SSLv23_method();
	if (ssl_sysdata.method == NULL)
		return ENODEV;

	ssl_sysdata.context = SSL_CTX_new (ssl_sysdata.method);
	if (ssl_sysdata.context == NULL)
		return ENODEV;
	if (cafile != NULL) {
		rc = SSL_CTX_load_verify_locations (ssl_sysdata.context, cafile, NULL);
		if (rc != 1) {
			SSL_CTX_free (ssl_sysdata.context);
			return EINVAL;
		}
	}
	rc = SSL_CTX_use_certificate_chain_file (ssl_sysdata.context, certfile);
	if (rc != 1) {
		SSL_CTX_free (ssl_sysdata.context);
		return EINVAL;
	}
	rc = SSL_CTX_use_PrivateKey_file (ssl_sysdata.context,
					keyfile,
					SSL_FILETYPE_PEM);
	if (rc != 1) {
		SSL_CTX_free (ssl_sysdata.context);
		return EINVAL;
	}
	if (cafile != NULL) {
		SSL_CTX_set_verify (ssl_sysdata.context,
				SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
				NULL);
		SSL_CTX_set_verify_depth (ssl_sysdata.context, 4);
	}
	SSL_CTX_set_options (ssl_sysdata.context, SSL_PROTOCOL_OPTIONS);
	rc = SSL_CTX_set_cipher_list (ssl_sysdata.context, SSL_CIPHER_LIST);
	if (rc != 1) {
		SSL_CTX_free (ssl_sysdata.context);
		return EINVAL;
	}

	cmd->name = "openssl";
	cmd->data = &ssl_sysdata;
	cmd->vector->bind = openssl_comm_bind;
	cmd->vector->accept = openssl_comm_accept;
	cmd->vector->accept_continue = openssl_comm_accept_continue;
	cmd->vector->connect = openssl_comm_connect;
	cmd->vector->read = openssl_comm_read;
	cmd->vector->write = openssl_comm_write;
	cmd->vector->close = openssl_comm_close;
	cmd->vector->get_fd = openssl_comm_get_fd;
	return 0;
}


int comm_cleanup (CommModuleData *cmd)
{
	SslSysData *ssd = NULL;
	if (cmd == NULL || cmd->data == NULL)
		return EINVAL;
	ssd = (SslSysData *)cmd->data;
	SSL_CTX_free (ssd->context);
	return 0;
}



int openssl_comm_bind(CommChannel *cc, SList *interfacelist, int port)
{
	int rc;
	int sock;
	int socketopt;
	struct sockaddr_in sa;
	SslChannelData *scd = NULL;
	/*int *ip = NULL; */
	if (cc == NULL || port < 1)
		return EINVAL;
	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		rc = errno;
		return rc;
	}
	memset (&sa, 0, sizeof (struct sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = htonl (INADDR_ANY);
	socketopt = 1;
	setsockopt (sock,
		SOL_SOCKET,
		SO_REUSEADDR,
		&socketopt,
		sizeof (socketopt));
	rc = bind (sock,
		(struct sockaddr *)&sa,
		sizeof (struct sockaddr_in));
	if (rc == -1) {
		rc = errno;
		close (sock);
		return rc;
	}
	rc = listen (sock, 5);
	if (rc == -1) {
		rc = errno;
		close (sock);
		return rc;
	}
	
	scd = calloc (1, sizeof (SslChannelData));
	if (scd == NULL) {
		close (sock);
		return ENOMEM;
	}
	scd->ssl = NULL;
	scd->fd = sock;
	cc->data = scd;
/*
	ip = calloc (1, sizeof (int));
	if (ip == NULL) {
		close (sock);
		return ENOMEM;
	}
	*ip = sock;
	cc->data = ip;
*/
	cc->flags |= COMMFLAG_SERVER;
	return 0;
}



int openssl_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags)
{
	int rc;
	int sock;
	int csock;
	int csalen;
	struct sockaddr_in csa;
	SslSysData *ssd = NULL;
	SslChannelData *scd = NULL;
	if (cc == NULL || ncc == NULL)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_SERVER))
		return EINVAL;
	scd = cc->data;
	sock = scd->fd;
/*	sock = *(int *)cc->data; */
	if (sock == -1)
		return EINVAL;
	csalen = sizeof (struct sockaddr);
	csock = accept (sock,
			(struct sockaddr *)&csa,
			&csalen);
	if (csock == -1) {
		return errno;
	}

	/*
	 * SSL_accept() could block, so set the new socket non-blocking now
	 * Just set O_NONBLOCK (only valid socket flag)
	*/
	if (! (cc->flags & COMMFLAG_IO_BLOCK)) {
		rc = fcntl (csock, F_SETFL, O_NONBLOCK);
		if (rc != 0) {
			rc = errno;
			close (csock);
			return rc;
		}
	}

	scd = calloc (1, sizeof (SslChannelData));
	if (scd == NULL) {
		close (csock);
		return ENOMEM;
	}
	ssd = (SslSysData *)cc->cmodule->moddata->data;
	scd->ssl = SSL_new (ssd->context);
	if (scd->ssl == NULL) {
		close (csock);
		free (scd);
		return (int)ERR_get_error();
	}
	scd->fd = csock;
	rc = SSL_set_fd (scd->ssl, scd->fd);
	if (rc != 1) {
		rc = SSL_get_error (scd->ssl, rc);
		SSL_free (scd->ssl);
		free (scd);
		close (csock);
		return rc;
	}
	ncc->data = scd;
	rc = SSL_accept (scd->ssl);
	if (rc != 1) {
		rc = SSL_get_error (scd->ssl, rc);
		if (rc == SSL_ERROR_WANT_READ || rc == SSL_ERROR_WANT_WRITE) {
			ncc->flags |= COMMFLAG_INACCEPT;
			return rc;
		}
		SSL_free (scd->ssl);
		free (scd);
		close (csock);
		return rc;
	}
	ncc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int openssl_comm_accept_continue(CommChannel *cc)
{
	int rc;
	int csock;
	SslChannelData *scd = NULL;
	if (cc == NULL)
		return EINVAL;
	if (cc->data == NULL)
		return EINVAL;
	scd = cc->data;
	rc = SSL_accept (scd->ssl);
	if (rc != 1) {
		rc = SSL_get_error (scd->ssl, rc);
		if (rc == SSL_ERROR_WANT_READ || rc == SSL_ERROR_WANT_WRITE)
			return rc;
		csock = SSL_get_fd (scd->ssl);
		SSL_free (scd->ssl);
		free (scd);
		close (csock);
	} else { 
		cc->flags -= (cc->flags & COMMFLAG_INACCEPT);
		cc->flags |= COMMFLAG_CONNECTED;
		rc = 0;
	}
	return rc;
}


int openssl_comm_connect(CommChannel *cc, char *host, int port)
{
	int rc;
	int sock;
	struct sockaddr_in sa;
	struct hostent *he = NULL;
	SslSysData *ssd = NULL;
	SslChannelData *scd = NULL;
	if (cc == NULL || host == NULL || port < 1)
		return EINVAL;
	he = gethostbyname (host);
	if (he == NULL)
		return ENOENT;
	memset (&sa, 0, sizeof (struct sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons (port);
	sa.sin_addr.s_addr = *(long *)he->h_addr_list[0];
	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		rc = errno;
		return rc;
	}
	rc = connect (sock,
			(struct sockaddr *)&sa,
			sizeof (struct sockaddr_in));
	if (rc == -1) {
		rc = errno;
		close (sock);
		return rc;
	}
	scd = calloc (1, sizeof (SslChannelData));
	if (scd == NULL) {
		close (sock);
		return ENOMEM;
	}
	ssd = (SslSysData *)cc->cmodule->moddata->data;
	scd->ssl = SSL_new (ssd->context);
	scd->fd = sock;
	rc = SSL_set_fd (scd->ssl, scd->fd);
	if (rc != 1) {
		rc = SSL_get_error (scd->ssl, rc);
		SSL_free (scd->ssl);
		free (scd);
		close (sock);
		return rc;
	}
	rc = SSL_connect (scd->ssl);
	if (rc != 1) {
		rc = SSL_get_error (scd->ssl, rc);
		SSL_free (scd->ssl);
		free (scd);
		close (sock);
		return rc;
	}
	cc->data = scd;
	cc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int openssl_comm_read(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int len;
	int bytesleft;
	char *p = NULL;
	SslChannelData *scd = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->flags & COMMFLAG_SERVER)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return EINVAL;
	scd = (SslChannelData *)cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = SSL_read (scd->ssl, p, bytesleft);
			if (rc <= 0) {
				*bytes = len - bytesleft;
				rc = SSL_get_error (scd->ssl, rc);
				if (rc == SSL_ERROR_SYSCALL)
					return errno;
				return rc;
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = SSL_read (scd->ssl, buffer, len);
		if (rc < 0) {
			rc = SSL_get_error (scd->ssl, rc);
			if (rc == SSL_ERROR_SYSCALL)
				return errno;
			return rc;
		}
		if (rc == 0)
			return ENODATA;
		*bytes = rc;
	}
	return 0;
}


int openssl_comm_write(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int len;
	int bytesleft;
	char *p = NULL;
	SslChannelData *scd = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->flags & COMMFLAG_SERVER)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return EINVAL;
	scd = (SslChannelData *)cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = SSL_write (scd->ssl, p, bytesleft);
			if (rc <= 0) {
				*bytes = len - bytesleft;
				rc = SSL_get_error (scd->ssl, rc);
				if (rc == SSL_ERROR_SYSCALL)
					return errno;
				return rc;
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = SSL_write (scd->ssl, buffer, len);
		if (rc <= 0) {
			rc = SSL_get_error (scd->ssl, rc);
			if (rc == SSL_ERROR_SYSCALL)
				return errno;
			return rc;
		}
		*bytes = rc;
	}
	return 0;
}


int openssl_comm_close(CommChannel *cc, int commflags)
{
	int rc;
	int sock;
	SslChannelData *scd = NULL;
	if (cc == NULL || cc->data == NULL)
		return EINVAL;
	sock = -1;
	if (cc->flags & COMMFLAG_SERVER) {
		scd = (SslChannelData *)cc->data;
		sock = scd->fd;
		/* sock = *(int *)cc->data; */
		close (sock);
		free (scd);
	} else {
		scd = (SslChannelData *)cc->data;
		if (scd == NULL || scd->ssl == NULL || scd->fd == -1)
			return EINVAL;
		rc = SSL_shutdown (scd->ssl);
		/* Do SSL 2-step shutdown, if needed */
		if (rc == 0)
			rc = SSL_shutdown (scd->ssl);
		SSL_free (scd->ssl);
		rc = close (scd->fd);
		free (scd);
	}
	return 0;
}


int openssl_comm_get_fd(CommChannel *cc, int *fdp)
{
	int		fd;
	SslChannelData	*scd = NULL;
	if (cc == NULL || fdp == NULL)
		return EINVAL;
	*fdp = -1;
	scd = (SslChannelData *)cc->data;
	if (cc->flags & COMMFLAG_SERVER) {
		*fdp = scd->fd;
	} else {
		fd = SSL_get_fd (scd->ssl);
		if (fd == -1)
			return SSL_get_error (scd->ssl, fd);
		*fdp = fd;
	}
	return 0;
}



