/************************************************************************/
/* comm_tcp.c - Communication Plugin Source (TCP)			*/
/************************************************************************/
/*
 *  (c)2003 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

#include <cutil/cutil.h>
#include <comm/comm.h>
#include <comm/comm_portable.h>


#ifndef WIN32
#define	SOCKET_ERRNO	errno
#else
#define	SOCKET_ERRNO	WSAGetLastError()
#endif


/* Function Prototypes */
int tcp_comm_bind(CommChannel *cc, SList *interfacelist, int port);
int tcp_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags);
int tcp_comm_accept_continue(CommChannel *cc);
int tcp_comm_connect(CommChannel *cc, char *host, int port);
int tcp_comm_read(CommChannel *cc, char *buffer, int *bytes);
int tcp_comm_write(CommChannel *cc, char *buffer, int *bytes);
int tcp_comm_close(CommChannel *cc, int commflags);
int tcp_comm_get_fd(CommChannel *cc, int *fdp);


/* Global variables */



int comm_init (CommModuleData *cmd)
{
	if (cmd == NULL)
		return EINVAL;
#ifdef WIN32
	{
	int		rc;
	WSADATA		wsadata;
	rc = WSAStartup (MAKEWORD(2,0), &wsadata);
	if (rc != 0)
		return rc;
	}
#endif
	cmd->name = "tcp";
	cmd->vector->bind = tcp_comm_bind;
	cmd->vector->accept = tcp_comm_accept;
	cmd->vector->accept_continue = tcp_comm_accept_continue;
	cmd->vector->connect = tcp_comm_connect;
	cmd->vector->read = tcp_comm_read;
	cmd->vector->write = tcp_comm_write;
	cmd->vector->close = tcp_comm_close;
	cmd->vector->get_fd = tcp_comm_get_fd;
	return 0;
}


int comm_cleanup (CommModuleData *cmd)
{
	if (cmd == NULL)
		return EINVAL;
	SOCKET_CLEANUP();
	return 0;
}


int tcp_comm_bind(CommChannel *cc, SList *interfacelist, int port)
{
	int rc;
	int sock;
	int socketopt;
	struct sockaddr_in sa;
	int *ip = NULL;
	if (cc == NULL || port < 1)
		return EINVAL;
	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		rc = SOCKET_ERRNO;
		return rc;
	}
	memset (&sa, 0, sizeof (struct sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons (port);
	sa.sin_addr.s_addr = htonl (INADDR_ANY);
	socketopt = 1;
	setsockopt (sock,
		SOL_SOCKET,
		SO_REUSEADDR,
		&socketopt,
		sizeof (socketopt));
	rc = bind (sock,
		(struct sockaddr *)&sa,
		sizeof (struct sockaddr_in));
	if (rc == -1) {
		rc = SOCKET_ERRNO;
		close (sock);
		return rc;
	}
	rc = listen (sock, 5);
	if (rc == -1) {
		rc = SOCKET_ERRNO;
		close (sock);
		return rc;
	}
	ip = calloc (1, sizeof (int));
	if (ip == NULL) {
		close (sock);
		return ENOMEM;
	}
	*ip = sock;
	cc->data = ip;
	cc->flags |= COMMFLAG_SERVER;
	return 0;
}



int tcp_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags)
{
	int sock;
	int csock;
	int csalen;
	struct sockaddr_in csa;
	int *ip = NULL;
	if (cc == NULL || ncc == NULL)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_SERVER))
		return EINVAL;
	sock = *(int *)cc->data;
	if (sock == -1)
		return EINVAL;
	csock = accept (sock,
			(struct sockaddr *)&csa,
			&csalen);
	if (csock == -1)
		return SOCKET_ERRNO;
	ip = calloc (1, sizeof (int));
	if (ip == NULL) {
		close (csock);
		return ENOMEM;
	}
	*ip = csock;
	ncc->data = ip;
	ncc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int tcp_comm_accept_continue(CommChannel *cc)
{
	if (cc == NULL)
		return EINVAL;
	return 0;
}


int tcp_comm_connect(CommChannel *cc, char *host, int port)
{
	int rc;
	int sock;
	struct sockaddr_in sa;
	int *ip = NULL;
	struct hostent *he = NULL;
	if (cc == NULL || host == NULL || port < 1)
		return EINVAL;
	he = gethostbyname (host);
	if (he == NULL)
		return ENOENT;
	memset (&sa, 0, sizeof (struct sockaddr_in));
	sa.sin_family = AF_INET;
	sa.sin_port = htons (port);
	sa.sin_addr.s_addr = *(long *)he->h_addr_list[0];
	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		rc = SOCKET_ERRNO;
		return rc;
	}
	rc = connect (sock,
			(struct sockaddr *)&sa,
			sizeof (struct sockaddr_in));
	if (rc == -1) {
		rc = SOCKET_ERRNO;
		close (sock);
		return rc;
	}
	ip = calloc (1, sizeof (int));
	if (ip == NULL) {
		close (sock);
		return ENOMEM;
	}
	*ip = sock;
	cc->data = ip;
	cc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int tcp_comm_read(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int sock;
	int len;
	int bytesleft;
	char *p = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->flags & COMMFLAG_SERVER)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return EINVAL;
	sock = *(int *)cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = recv (sock, p, bytesleft, 0);
			if (rc == -1) {
				*bytes = len - bytesleft;
				return SOCKET_ERRNO;
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = recv (sock, buffer, len, 0);
		if (rc == -1)
			return SOCKET_ERRNO;
		if (rc == 0)
			return ENODATA;
		*bytes = rc;
	}
	return 0;
}


int tcp_comm_write(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int sock;
	int len;
	int bytesleft;
	char *p = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->flags & COMMFLAG_SERVER)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return EINVAL;
	sock = *(int *)cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = send (sock, p, bytesleft, 0);
			if (rc == -1) {
				*bytes = len - bytesleft;
				return SOCKET_ERRNO;
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = send (sock, buffer, len, 0);
		if (rc == -1)
			return SOCKET_ERRNO;
		*bytes = rc;
	}
	return 0;
}


int tcp_comm_close(CommChannel *cc, int commflags)
{
	int rc;
	int sock;
	if (cc == NULL)
		return EINVAL;
	sock = *(int *)cc->data;
	if (sock == -1)
		return EINVAL;
	rc = close (sock);
	free (cc->data);
	return 0;
}


int tcp_comm_get_fd(CommChannel *cc, int *fdp)
{
	if (cc == NULL || fdp == NULL)
		return EINVAL;
	*fdp = *(int *)cc->data;
	return 0;
}



