/************************************************************************/
/* comm_unix.c - Communication Plugin Source (UNIX)			*/
/************************************************************************/
/*
 *  (c)2007 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <cutil/cutil.h>
#include <comm/comm.h>
#include <comm/comm_portable.h>


typedef struct _UnixChannelData	UnixChannelData;
struct _UnixChannelData {
	int	fd;
	char	*sockpath;
};


/* Function Prototypes */
int unix_comm_bind(CommChannel *cc, SList *interfacelist, int port);
int unix_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags);
int unix_comm_accept_continue(CommChannel *cc);
int unix_comm_connect(CommChannel *cc, char *host, int port);
int unix_comm_read(CommChannel *cc, char *buffer, int *bytes);
int unix_comm_write(CommChannel *cc, char *buffer, int *bytes);
int unix_comm_close(CommChannel *cc, int commflags);
int unix_comm_get_fd(CommChannel *cc, int *fdp);


/* Global variables */



int comm_init (CommModuleData *cmd)
{
	if (cmd == NULL)
		return EINVAL;
	cmd->name = "unix";
	cmd->vector->bind = unix_comm_bind;
	cmd->vector->accept = unix_comm_accept;
	cmd->vector->accept_continue = unix_comm_accept_continue;
	cmd->vector->connect = unix_comm_connect;
	cmd->vector->read = unix_comm_read;
	cmd->vector->write = unix_comm_write;
	cmd->vector->close = unix_comm_close;
	cmd->vector->get_fd = unix_comm_get_fd;
	return 0;
}


int comm_cleanup (CommModuleData *cmd)
{
	if (cmd == NULL)
		return EINVAL;
	return 0;
}


int unix_comm_bind(CommChannel *cc, SList *interfacelist, int port)
{
	int rc;
	int sock;
	int socketopt;
	struct sockaddr_un sa;
	char *sockpath = NULL;
	UnixChannelData	*ucd = NULL;
	if (cc == NULL || interfacelist == NULL)
		return EINVAL;
	sockpath = slist_get_data (interfacelist);
	if (sockpath == NULL)
		return EINVAL;
	ucd = calloc (1, sizeof (UnixChannelData));
	if (ucd == NULL)
		return ENOMEM;
	ucd->sockpath = strdup (sockpath);
	if (ucd->sockpath == NULL) {
		free (ucd);
		return ENOMEM;
	}
	sock = socket (PF_UNIX, SOCK_STREAM, 0);
	if (sock == -1) {
		rc = errno;
		return rc;
	}
	memset (&sa, 0, sizeof (struct sockaddr_un));
	sa.sun_family = AF_UNIX;
	strncpy (sa.sun_path, ucd->sockpath, strlen (ucd->sockpath));
	socketopt = 1;
	setsockopt (sock,
		SOL_SOCKET,
		SO_REUSEADDR,
		&socketopt,
		sizeof (socketopt));
	rc = bind (sock,
		(struct sockaddr *)&sa,
		sizeof (struct sockaddr_un));
	if (rc == -1) {
		rc = errno;
		close (sock);
		return rc;
	}
	rc = listen (sock, 5);
	if (rc == -1) {
		rc = errno;
		close (sock);
		return rc;
	}
	ucd->fd = sock;
	cc->data = ucd;
	cc->flags |= COMMFLAG_SERVER;
	return 0;
}



int unix_comm_accept(CommChannel *cc, CommChannel *ncc, int commflags)
{
	int sock;
	int csock;
	int csalen;
	struct sockaddr_un csa;
	UnixChannelData	*ucd = NULL;
	UnixChannelData	*ucdp = NULL;
	if (cc == NULL || ncc == NULL)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_SERVER))
		return EINVAL;
	sock = *(int *)cc->data;
	if (sock == -1)
		return EINVAL;
	csock = accept (sock,
			(struct sockaddr *)&csa,
			&csalen);
	if (csock == -1)
		return errno;
	ucd = calloc (1, sizeof (UnixChannelData));
	if (ucd == NULL) {
		close (csock);
		return ENOMEM;
	}
	ucdp = cc->data;
	ucd->sockpath = strdup (ucdp->sockpath);
	if (ucd->sockpath == NULL) {
		close (csock);
		free (ucd);
		return ENOMEM;
	}
	ucd->fd = csock;
	ncc->data = ucd;
	ncc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int unix_comm_accept_continue(CommChannel *cc)
{
	if (cc == NULL)
		return EINVAL;
	return 0;
}


int unix_comm_connect(CommChannel *cc, char *host, int port)
{
	int rc;
	int sock;
	struct sockaddr_un sa;
	UnixChannelData	*ucd = NULL;
	if (cc == NULL || host == NULL)
		return EINVAL;
	ucd = calloc (1, sizeof (UnixChannelData));
	if (ucd == NULL)
		return ENOMEM;
	ucd->sockpath = strdup (host);
	if (ucd->sockpath == NULL) {
		free (ucd);
		return ENOMEM;
	}
	memset (&sa, 0, sizeof (struct sockaddr_un));
	sa.sun_family = AF_UNIX;
	strncpy (sa.sun_path, ucd->sockpath, strlen(ucd->sockpath));
	sock = socket (PF_UNIX, SOCK_STREAM, 0);
	if (sock == -1) {
		rc = errno;
		return rc;
	}
	rc = connect (sock,
			(struct sockaddr *)&sa,
			sizeof (struct sockaddr_un));
	if (rc == -1) {
		rc = errno;
		close (sock);
		return rc;
	}
	ucd->fd = sock;
	cc->data = ucd;
	cc->flags |= COMMFLAG_CONNECTED;
	return 0;
}


int unix_comm_read(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int sock;
	int len;
	int bytesleft;
	char *p = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->flags & COMMFLAG_SERVER)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return EINVAL;
	sock = *(int *)cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = read (sock, p, bytesleft);
			if (rc == -1) {
				*bytes = len - bytesleft;
				return errno;
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = read (sock, buffer, len);
		if (rc == -1)
			return errno;
		if (rc == 0)
			return ENODATA;
		*bytes = rc;
	}
	return 0;
}


int unix_comm_write(CommChannel *cc, char *buffer, int *bytes)
{
	int rc;
	int sock;
	int len;
	int bytesleft;
	char *p = NULL;
	if (cc == NULL || buffer == NULL || bytes == NULL)
		return EINVAL;
	if (cc->flags & COMMFLAG_SERVER)
		return EINVAL;
	if (!(cc->flags & COMMFLAG_CONNECTED))
		return EINVAL;
	sock = *(int *)cc->data;
	len = *bytes;
	if (cc->flags & COMMFLAG_IO_BLOCK) {
		bytesleft = len;
		p = buffer;
		while (bytesleft > 0) {
			rc = write (sock, p, bytesleft);
			if (rc == -1) {
				*bytes = len - bytesleft;
				return errno;
			}
			bytesleft -= rc;
			p += rc;
		}
		*bytes = len - bytesleft;
	} else {
		rc = write (sock, buffer, len);
		if (rc == -1)
			return errno;
		*bytes = rc;
	}
	return 0;
}


int unix_comm_close(CommChannel *cc, int commflags)
{
	int rc;
	UnixChannelData	*ucd = NULL;
	if (cc == NULL)
		return EINVAL;
	ucd = cc->data;
	if (ucd == NULL)
		return EINVAL;
	if (ucd->fd == -1)
		return EINVAL;
	rc = close (ucd->fd);
	if (cc->flags & COMMFLAG_SERVER)
		rc = unlink (ucd->sockpath);
	if (ucd->sockpath != NULL)
		free (ucd->sockpath);
	free (ucd);
	return 0;
}


int unix_comm_get_fd(CommChannel *cc, int *fdp)
{
	if (cc == NULL || fdp == NULL)
		return EINVAL;
	*fdp = *(int *)cc->data;
	return 0;
}



